use crate::environment::Txn;
use crate::wrapper::ZeroCopyByteWrapper;
use libmdbx::{Database, Environment, TransactionKind, WriteFlags, WriteMap, RW};
use ouroboros::self_referencing;
use std::borrow::Cow;
use std::marker::PhantomData;
use std::sync::Arc;

#[self_referencing(pub_extras)]
pub struct RawTable<K: ?Sized, V: ?Sized> {
    pub mdbx_env: Arc<Environment<WriteMap>>,

    #[borrows(mdbx_env)]
    #[covariant]
    pub mdbx_db: Database<'this>,

    pub(crate) phantom_k: PhantomData<K>,
    pub(crate) phantom_v: PhantomData<V>,
}

impl<K: ?Sized + ZeroCopyByteWrapper, V: ?Sized + ZeroCopyByteWrapper> RawTable<K, V> {
    pub fn get<'txn, TK: TransactionKind>(
        &'txn self,
        txn: &'txn Txn<'txn, TK>,
        k: impl AsRef<K>,
    ) -> anyhow::Result<Option<Cow<'txn, V>>> {
        let bytes = txn
            .mdbx_txn
            .get::<Cow<'txn, [u8]>>(self.borrow_mdbx_db(), k.as_ref().as_byte_slice())?;
        Ok(match bytes {
            None => None,
            Some(Cow::Owned(owned)) => Some(Cow::Owned(V::from_owned_bytes(owned)?)),
            Some(Cow::Borrowed(borrowed)) => Some(Cow::Borrowed(V::from_byte_slice(borrowed)?)),
        })
    }

    pub fn put<'txn>(
        &self,
        txn: &Txn<'txn, RW>,
        k: impl AsRef<K>,
        v: impl AsRef<V>,
    ) -> anyhow::Result<()> {
        txn.mdbx_txn.put(
            self.borrow_mdbx_db(),
            k.as_ref().as_byte_slice(),
            v.as_ref().as_byte_slice(),
            WriteFlags::empty(),
        )?;
        Ok(())
    }

    /// Returns true if the key was present.
    pub fn delete<'txn>(&self, txn: &Txn<'txn, RW>, k: impl AsRef<K>) -> anyhow::Result<bool> {
        Ok(txn
            .mdbx_txn
            .del(self.borrow_mdbx_db(), k.as_ref().as_byte_slice(), None)?)
    }
}
