use crate::database::raw::RawTable;
use crate::environment::Txn;
use crate::wrapper::ByteWrapper;
use anyhow::Context;
use libmdbx::{TransactionKind, RW};
use std::borrow::Borrow;

pub struct WrappedTable<K, V> {
    pub raw: RawTable<[u8], [u8]>,
    pub k_wrapper: K,
    pub v_wrapper: V,
}

impl<K: ByteWrapper, V: ByteWrapper> WrappedTable<K, V> {
    pub fn get<'txn, TK: TransactionKind>(
        &self,
        txn: &Txn<'txn, TK>,
        k: impl Borrow<K::Item>,
    ) -> anyhow::Result<Option<V::Item>> {
        let k_bytes = self
            .k_wrapper
            .dump_to_db_bytes(k.borrow())
            .context("whilst converting key to bytes")?;

        self.raw
            .get(txn, k_bytes)?
            .map(|v_bytes| self.v_wrapper.load_from_db_bytes(&v_bytes))
            .transpose()
    }

    pub fn put<'txn>(
        &self,
        txn: &Txn<'txn, RW>,
        k: impl Borrow<K::Item>,
        v: impl Borrow<V::Item>,
    ) -> anyhow::Result<()> {
        let k_bytes = self
            .k_wrapper
            .dump_to_db_bytes(k.borrow())
            .context("whilst converting key to bytes")?;
        let v_bytes = self
            .v_wrapper
            .dump_to_db_bytes(v.borrow())
            .context("whilst converting value to bytes")?;

        self.raw.put(txn, k_bytes, v_bytes)?;
        Ok(())
    }

    /// Returns true if the key was present.
    pub fn delete<'txn>(
        &self,
        txn: &Txn<'txn, RW>,
        k: impl Borrow<K::Item>,
    ) -> anyhow::Result<bool> {
        let k_bytes = self
            .k_wrapper
            .dump_to_db_bytes(k.borrow())
            .context("whilst converting key to bytes")?;

        self.raw.delete(txn, k_bytes)
    }
}
