use crate::database::raw::{RawTable, RawTableTryBuilder};
use crate::database::wrapped::WrappedTable;
use crate::wrapper::{ByteWrapper, ZeroCopyByteWrapper};
use anyhow::{ensure, Context};
use libmdbx::{
    DatabaseFlags, Environment, EnvironmentFlags, Geometry, Transaction, TransactionKind, WriteMap,
    RO, RW,
};
use std::marker::PhantomData;
use std::path::Path;
use std::sync::Arc;

pub struct Env {
    pub mdbx_env: Arc<Environment<WriteMap>>,
}

impl Env {
    pub fn open(path: &Path) -> anyhow::Result<Env> {
        let mut flags = EnvironmentFlags::default();
        flags.no_sub_dir = true;

        // TODO make the geometry more configurable.
        let mut geom = Geometry::default();
        // Don't stop the database growing until it hits 64 GiB.
        // (The default is 1 MiB which is just not enough!)
        geom.size = Some(1024 * 1024..64 * 1024 * 1024 * 1024);

        // Grow 4 MiB at a time.
        geom.growth_step = Some(4 * 1024 * 1024);
        // Shrink 16 MiB at a time.
        geom.shrink_threshold = Some(16 * 1024 * 1024);
        // (Yes these numbers represent a large database).

        let environment = Environment::new()
            .set_max_dbs(256)
            .set_geometry(geom)
            .set_flags(flags)
            .open(path)?;

        Ok(Env {
            mdbx_env: Arc::new(environment),
        })
    }

    pub fn open_raw_table<K: ZeroCopyByteWrapper + ?Sized, V: ZeroCopyByteWrapper + ?Sized>(
        &self,
        name: Option<&str>,
        _flags: (),
    ) -> anyhow::Result<RawTable<K, V>> {
        Ok(RawTableTryBuilder {
            mdbx_env: self.mdbx_env.clone(),
            mdbx_db_builder: |mdbx_env: &Arc<Environment<WriteMap>>| {
                let txn = mdbx_env
                    .begin_rw_txn()
                    .context("Can't start RW transaction")?;
                // TODO database flags
                let db = txn
                    .create_db(name, DatabaseFlags::empty())
                    .context("Can't open database")?;
                txn.prime_for_permaopen(db);

                let (_bool, mut dbs) = txn.commit_and_rebind_open_dbs()?;

                ensure!(dbs.len() == 1);

                let mdbx_db = dbs.pop().context("Expected database!")?;

                Ok(mdbx_db)
            },
            phantom_k: PhantomData::default(),
            phantom_v: PhantomData::default(),
        }
        .try_build()?)
    }

    pub fn open_wrapped_table<K: ByteWrapper, V: ByteWrapper>(
        &self,
        name: Option<&str>,
        flags: (),
        k_wrapper: K,
        v_wrapper: V,
    ) -> anyhow::Result<WrappedTable<K, V>> {
        let raw_table = self.open_raw_table(name, flags)?;

        Ok(WrappedTable {
            raw: raw_table,
            k_wrapper,
            v_wrapper,
        })
    }

    /// Begins a read-only transaction.
    pub fn ro_txn(&self) -> anyhow::Result<Txn<RO>> {
        let mdbx_txn = self.mdbx_env.begin_ro_txn()?;
        Ok(Txn { mdbx_txn })
    }

    /// Begins a read-write transaction. This one has an awkward name because `rw_txn` is an easier
    /// interface (insofar as you don't need to remember to `commit` yourself).
    pub fn begin_rw_txn(&self) -> anyhow::Result<Txn<RW>> {
        let mdbx_txn = self.mdbx_env.begin_rw_txn()?;
        Ok(Txn { mdbx_txn })
    }

    /// Runs a transaction and commits for you, if the Result variant is `Ok(R)` (NOT `Err`).
    /// This way, it's hard to forget to commit.
    pub fn rw_txn<R>(&self, func: impl FnOnce(&Txn<RW>) -> anyhow::Result<R>) -> anyhow::Result<R> {
        let txn = self.begin_rw_txn()?;
        let result = func(&txn)?;
        txn.commit().context("Failed to commit")?;
        Ok(result)
    }
}

pub struct Txn<'txn, K: TransactionKind> {
    pub mdbx_txn: Transaction<'txn, K, WriteMap>,
}

impl<'txn, K: TransactionKind> Txn<'txn, K> {}

impl<'txn> Txn<'txn, RW> {
    pub fn commit(self) -> anyhow::Result<()> {
        // TODO bool here, find out what it means and maybe pass up
        self.mdbx_txn.commit()?;
        Ok(())
    }
}
