use anyhow::anyhow;
use serde::de::DeserializeOwned;
use serde::Serialize;
use std::marker::PhantomData;
use std::sync::{Arc, Mutex};
use zstd::bulk::{Compressor, Decompressor};

pub trait ByteWrapper {
    type Item;

    fn load_from_db_bytes(&self, bytes: &[u8]) -> anyhow::Result<Self::Item>;

    fn dump_to_db_bytes(&self, item: &Self::Item) -> anyhow::Result<Vec<u8>>;
}

pub struct SerdeBareWrapper<T> {
    phantom: PhantomData<T>,
}

impl<T> Default for SerdeBareWrapper<T> {
    fn default() -> Self {
        SerdeBareWrapper {
            phantom: Default::default(),
        }
    }
}

impl<T: Serialize + DeserializeOwned> ByteWrapper for SerdeBareWrapper<T> {
    type Item = T;

    fn load_from_db_bytes(&self, bytes: &[u8]) -> anyhow::Result<Self::Item> {
        Ok(serde_bare::from_slice(bytes)?)
    }

    fn dump_to_db_bytes(&self, item: &Self::Item) -> anyhow::Result<Vec<u8>> {
        Ok(serde_bare::to_vec(item)?)
    }
}

pub struct CompressorWrapper<T> {
    inner: T,
    compressor: Mutex<Compressor<'static>>,
    decompressor: Mutex<Decompressor<'static>>,
}

impl<T: ByteWrapper> ByteWrapper for CompressorWrapper<T> {
    type Item = T::Item;

    fn load_from_db_bytes(&self, bytes: &[u8]) -> anyhow::Result<Self::Item> {
        let mut decompressor = self
            .decompressor
            .lock()
            .map_err(|_| anyhow!("Can't lock decompressor"))?;
        // TODO be more sensible about capacity. 8× should be good though.
        let raw_bytes = decompressor.decompress(bytes, bytes.len() * 8)?;
        Ok(self.inner.load_from_db_bytes(&raw_bytes)?)
    }

    fn dump_to_db_bytes(&self, item: &Self::Item) -> anyhow::Result<Vec<u8>> {
        let raw_bytes = self.inner.dump_to_db_bytes(item)?;
        let mut compressor = self
            .compressor
            .lock()
            .map_err(|_| anyhow!("Can't lock compressor"))?;
        Ok(compressor.compress(&raw_bytes)?)
    }
}

impl<T> CompressorWrapper<T> {
    pub fn new_with(inner: T) -> anyhow::Result<Self> {
        let compressor = Compressor::new(13)?;
        let decompressor = Decompressor::new()?;

        Ok(CompressorWrapper {
            inner,
            compressor: Mutex::new(compressor),
            decompressor: Mutex::new(decompressor),
        })
    }

    pub fn new() -> anyhow::Result<Self>
    where
        T: Default,
    {
        let inner: T = Default::default();

        Self::new_with(inner)
    }
}

impl ByteWrapper for String {
    type Item = String;

    fn load_from_db_bytes(&self, bytes: &[u8]) -> anyhow::Result<Self::Item> {
        Ok(std::str::from_utf8(bytes)?.to_string())
    }

    fn dump_to_db_bytes(&self, item: &Self::Item) -> anyhow::Result<Vec<u8>> {
        Ok(item.as_bytes().to_vec())
    }
}

impl<T: ByteWrapper> ByteWrapper for Arc<T> {
    type Item = T::Item;

    fn load_from_db_bytes(&self, bytes: &[u8]) -> anyhow::Result<Self::Item> {
        (&self as &T).load_from_db_bytes(bytes)
    }

    fn dump_to_db_bytes(&self, item: &Self::Item) -> anyhow::Result<Vec<u8>> {
        (&self as &T).dump_to_db_bytes(item)
    }
}

pub trait ZeroCopyByteWrapper: ToOwned {
    fn as_byte_slice(&self) -> &[u8];
    fn from_byte_slice(bytes: &[u8]) -> anyhow::Result<&Self>;
    fn from_owned_bytes(bytes: Vec<u8>) -> anyhow::Result<Self::Owned>;
}

impl ZeroCopyByteWrapper for [u8] {
    fn as_byte_slice(&self) -> &[u8] {
        self
    }

    fn from_byte_slice(bytes: &[u8]) -> anyhow::Result<&Self> {
        Ok(bytes)
    }

    fn from_owned_bytes(bytes: Vec<u8>) -> anyhow::Result<Self::Owned> {
        Ok(bytes)
    }
}

impl ZeroCopyByteWrapper for str {
    fn as_byte_slice(&self) -> &[u8] {
        self.as_bytes()
    }

    fn from_byte_slice(bytes: &[u8]) -> anyhow::Result<&Self> {
        Ok(std::str::from_utf8(bytes)?)
    }

    fn from_owned_bytes(bytes: Vec<u8>) -> anyhow::Result<Self::Owned> {
        Ok(String::from_utf8(bytes)?)
    }
}
